from http.server import BaseHTTPRequestHandler, HTTPServer
import time
import json
import os
import pathlib
import subprocess

hostName = "127.0.0.1"
serverPort = 8080


def parse_time(line):
    time = line.split("\t")[-1]
    mins, secs = time.split("m")
    minutes = int(mins)
    seconds = float(secs[:-1].replace(",","."))
    return minutes*60+seconds


def parse_elapsed_time(stderr):
    elapsed = 0
    for line in stderr.decode("utf-8").split("\n")[-3:]:
        if line.startswith("user"):
            elapsed += parse_time(line)
        elif line.startswith("sys"):
            elapsed += parse_time(line)
    return elapsed

def parse_tamarin_out(result):
    is_in_results = False
    wait_results = False
    results_prop = []
    lemma_names = []
    for line in result.decode("utf-8").split("\n"):
        if line.startswith("========") and not is_in_results:
            is_in_results = True
        elif line.startswith("======="):
            is_in_results = False
        elif is_in_results:
            if "(all-traces)" in line or "(exists-trace)" in line:
                res = line.split(":")
                name = res[0].strip().split(" ")[0]
                if "verified" in res[1]:
                    results_prop.append(True)
                else:
                    results_prop.append(False)
                lemma_names.append(name)

    return (results_prop, lemma_names)

class ModelParser(object):
    def __init__(self, filename):
        self.filename = filename
        if not filename.endswith(".spthy"):
            raise ValueError("Expect a filename ending in .spthy")

        self.basename = filename[:-6]
        self.conf_file = self.basename+".json" # If we have a "model.spthy" file, expect a "model.json" file to launch the commands

        self.results_directory = os.path.join(".", "results", self.basename)
        if not os.path.exists(self.results_directory):
            os.makedirs(self.results_directory)

        with open(os.path.join("./models", self.conf_file)) as fp_conf:
            conf = json.load(fp_conf)

        self.properties = conf["properties"]
        self.clis = {}
        for name, flags in conf["submodels"].items():
            cli = ["tamarin-prover", os.path.join("./models", self.filename)]
            for flag in flags:
                if not flag.isalnum():
                    raise ValueError("Line {}: {} from the description of the model is not alphanumeric".format(name, flags))
                cli.append("-D{}".format(flag))
            self.clis[name] = cli

        self.attack_list = AttackList()
        self._init_attack_list()

    def _init_attack_list(self):
        results_path = pathlib.Path(f"results/{self.basename}")
        files = results_path.glob("*/result.spthy")
        res = {}
        for f in files:
            submodel_name = f.parent.name
            res[submodel_name] = f
            parser = Parser(open(f, "r").readlines())
            proofs = parser.get_proofs()
            for k, v in proofs.items():
                self.attack_list.add_trace(v, k)

    def get_name(self):
        return self.filename

    def get_basename(self):
        return self.basename

    def get_submodel_names(self):
        return list(sorted(self.clis.keys()))

    def get_submodel(self, submodel):
        return self.clis[submodel]

    def get_properties(self):
        return self.properties

    def execute_submodel(self, submodel):
        submodel_results = os.path.join(self.results_directory, submodel)
        if not os.path.exists(submodel_results):
            os.makedirs(submodel_results)
        filename = os.path.join(submodel_results, "result.spthy")
        summarized_filename = os.path.join(submodel_results, "result.json")
        if os.path.exists(summarized_filename):
            os.remove(summarized_filename)
        if os.path.exists(filename):
            os.remove(filename)
        
        
        cli = self.clis[submodel].copy()
        cli.append("--prove")
        cli.append("--output={}".format(filename))
        cli.append("2>/dev/null".format(filename))
        cli = ["timeout", "10m", "time"] + cli
        cli = " ".join(cli)

        result = subprocess.run(cli, capture_output=True, shell=True)
        elapsed_time = parse_elapsed_time(result.stderr)

        props_state, lemma_names = parse_tamarin_out(result.stdout)

        with open(summarized_filename, "w") as fp:
            d = {"submodel": submodel, "properties": props_state, "lemmas": lemma_names, "elapsed": elapsed_time}
            fp.write(json.dumps(d))

        return summarized_filename

    def get_results_file(self, submodel):
        submodel_results = os.path.join(self.results_directory, submodel)
        summarized_filename = os.path.join(submodel_results, "result.json")

        if os.path.exists(summarized_filename):
            return summarized_filename
        else:
            return None

    def get_repr_attack(self, submodel, idx_prop):
        res_file = self.get_results_file(submodel)
        with open(res_file, "r") as fp:
            content = fp.read()
            parsed = json.loads(content)
            lemma_status = parsed["properties"][idx_prop]
            if "lemmas" not in parsed:
                name = "V" if lemma_status else "X"
            elif not lemma_status:
                if "functional" in parsed["lemmas"][idx_prop]:
                    name = "X"
                else:
                    name = "X - " + self.attack_list.get_attack_name(parsed["lemmas"][idx_prop])
            else:
                name = "V"
            if lemma_status:
                return "<span class=\"proved\">"+name+"</span>"
            else:
                return "<span class=\"attack\">"+name+"</span>"


class AttackList(object):
    def __init__(self):
        self._attack_list = {}
        self._attack_names = {}
        self._ctr_attack = 1

    def add_trace(self, attack_trace, lemma_name):
        if not attack_trace in self._attack_list:
            self._attack_list[attack_trace] = []
            if "SOLVED // trace found" in attack_trace:
                self._attack_names[lemma_name] = "{}".format(self._ctr_attack)
                if not lemma_name.endswith("functional"):
                    self._ctr_attack += 1
        elif "SOLVED // trace found" in attack_trace:
            self._attack_names[lemma_name] = self._attack_names[self._attack_list[attack_trace][0]]
        self._attack_list[attack_trace].append(lemma_name)

    def get_attack_name(self, lemma):
        if not lemma in self._attack_names:
            return ""
        return self._attack_names[lemma]

class Parser(object):
    STATE_WAITING = 0
    STATE_START_PROOF = 1
    STATE_SIMPLIFY_PROOF = 2
    STATE_WAIT_BLANK = 3
    STATE_IN_PROOF = 4

    def __init__(self, lines):
        self._lines = lines
        self._curr_proof = []
        self._lemmas = {}
        self._state = self.STATE_WAITING
        self._lemma_list = []

        for line in lines:
            self.process_line(line)

    def get_lemma_name(self, lemma_line):
        a, b = lemma_line.strip().split(" ")
        b = b[:-1]
        return b

    def process_line(self, line):
        if self._state == self.STATE_WAITING:
            if line.startswith("lemma"):
                self._state = self.STATE_START_PROOF
                self._curr_proof.append(line)
        elif self._state == self.STATE_START_PROOF:
            if line.startswith("simplify"):
                self._state = self.STATE_SIMPLIFY_PROOF
                self._curr_proof.append(line)
        elif self._state == self.STATE_SIMPLIFY_PROOF:
            if line.startswith("by solve"):
                self._state = self.STATE_WAIT_BLANK
            else:
                self._state = self.STATE_IN_PROOF
            self._curr_proof.append(line)
        elif self._state == self.STATE_WAIT_BLANK:
            # Reach end of trace, store it
            if line.strip() == "":
                lemma_name = self.get_lemma_name(self._curr_proof[0])
                self._lemma_list.append(lemma_name)
                # We skip the lemma name
                self._lemmas[lemma_name] = "".join(self._curr_proof[1:])
                self._curr_proof = []
                self._state = self.STATE_WAITING
            else:
                self._curr_proof.append(line)
        elif self._state == self.STATE_IN_PROOF:
            if line.startswith("qed"):
                self._state = self.STATE_WAIT_BLANK
            self._curr_proof.append(line)

    def get_proofs(self):
        return self._lemmas

class MyServer(HTTPServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.models = []
        for filename in os.listdir("./models"):
            if filename.endswith(".spthy"):
                self.models.append(ModelParser(filename))

    def get_model(self, name):
        for model in self.models:
            if model.get_basename() == name:
                return model
        return None

class MyHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        params = self.path.strip("/").split("/")[3:]
        if len(params) == 0:
            self.request_home()
        elif len(params) == 1 and params[0] == "favicon.ico":
            self.request_404()
        elif len(params) == 1:
            self.request_model_index(params[0])
        elif len(params) == 2 and params[0] == "table":
            self.request_table(params[0], params[1])
        elif len(params) == 2:
            self.request_submodel(params[0], params[1])
        else:
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(bytes("<html><head><title>Tamarin output parser</title></head>", "utf-8"))
            self.wfile.write(bytes("<p>Request: {}</p>".format(self.path), "utf-8"))
            self.wfile.write(bytes("<body>", "utf-8"))
            self.wfile.write(bytes("<p>This is an example web server.</p>", "utf-8"))
            self.wfile.write(bytes("</body></html>", "utf-8"))

    def do_POST(self):
        params = self.path.strip("/").split("/")
        if len(params) == 3 and params[0] == "verify":
            self.prove_model(params[1], params[2])
        else:
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(bytes("<html><head><title>Tamarin output parser</title></head>", "utf-8"))
            self.wfile.write(bytes("<p>Request: {}</p>".format(self.path), "utf-8"))
            self.wfile.write(bytes("<body>", "utf-8"))
            self.wfile.write(bytes("<p>This is an example web server.</p>", "utf-8"))
            self.wfile.write(bytes("</body></html>", "utf-8"))

    def request_table(self, p1, p2):
        model_name = p2
        model = None
        for m in self.server.models:
            if m.get_basename() == model_name:
                model = m
        if model is None:
            raise ValueError("Model {} doesn't exist".format(model_name))

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes("<html><head><title>Latex table: {}</title>".format(model_name), "utf-8"))
        self.wfile.write(bytes("</head><body>", "utf-8"))
        self.wfile.write(bytes("<h2>{}</h2>".format(model_name), "utf-8"))

        self.wfile.write(bytes("<pre>", "utf-8"))
        self.wfile.write(bytes("""
\\begin{table}[]
\\begin{tabular}{|l|c|c|c|c|c|c|c|}
\\hline
Name""", "utf-8"))
        for prop in model.get_properties():
            self.wfile.write(bytes(" & {}".format(prop), "utf-8"))
        self.wfile.write(bytes(" & Time \\\\ \\hline\n", "utf-8"))
        for submodel in model.get_submodel_names():
            self.wfile.write(self.get_submodel_latex(model, submodel))
        self.wfile.write(bytes("""
\\end{tabular}
\\end{table}
</pre>""", "utf-8"))
        self.wfile.write(bytes("</script></body></html>", "utf-8"))


    def prove_model(self, model_name, submodel):
        model = self.server.get_model(model_name)
        results_fn = model.execute_submodel(submodel)

        fp = open(results_fn, "rb")
        content = fp.read()
        fp.close()

        self.send_response(200)
        self.send_header("Content-type", "application/javascript")
        self.end_headers()
        self.wfile.write(content)


    def request_home(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes("<html><head><title>Tamarin output parser</title></head>", "utf-8"))
        self.wfile.write(bytes("<p>Model list:</p>", "utf-8"))
        for model in self.server.models:
            self.wfile.write(bytes("<a href=\"/{}\">{}</a><br /><br />".format(model.get_basename(), model.get_name()), "utf-8"))

    def request_404(self):
        self.send_response(404)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes("<html><head><title>Not found</title></head></html>", "utf-8"))


    def request_model_index(self, model_name):
        model = None
        for m in self.server.models:
            if m.get_basename() == model_name:
                model = m
        if model is None:
            raise ValueError("Model {} doesn't exist".format(model_name))

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes("<html><head><title>Tamarin summary: {}</title>".format(model_name), "utf-8"))
        self.wfile.write(bytes("""<style>
body {
   padding: 2%;
   text-align: center;
}
th {
    text-align: center;
}
td,tr {
    border: 1px solid;
}
.proved {
   font-size: 18px;
   font-weight: bold;
   color: green;
}
.attack {
   font-size: 18px;
   font-weight: bold;
   color: red;
}
</style>""", "utf-8"))
        self.wfile.write(bytes("</head><body>".format(model_name), "utf-8"))
        self.wfile.write(bytes("<h2>{}</h2>".format(model_name), "utf-8"))

        self.wfile.write(bytes("<table>", "utf-8"))
        self.wfile.write(bytes("<tr>", "utf-8"))
        self.wfile.write(bytes("<th>Name</th>", "utf-8"))
        for prop in model.get_properties():
            self.wfile.write(bytes("<th>{}</th>".format(prop), "utf-8"))
        self.wfile.write(bytes("<th>Action</th>", "utf-8"))
        self.wfile.write(bytes("<th>Execution time</th>", "utf-8"))
        self.wfile.write(bytes("</tr>", "utf-8"))
        for submodel in model.get_submodel_names():
            self.wfile.write(self.get_submodel_html(model, submodel))
        self.wfile.write(bytes("</table>", "utf-8"))
        self.wfile.write(bytes("<br /><br /><a href=\"/table/{}\">Latex</a>".format(model_name), "utf-8"))

        self.wfile.write(bytes("""
    <script>
    function proveSubmodel(model_name, submodel) {
        var xhr = new XMLHttpRequest();
        xhr.open(\"POST\", \"/verify/\"+model_name+\"/\"+submodel);
        xhr.onreadystatechange = function() {
            if(xhr.readyState === 4 && xhr.status === 200) {
            reloadLine(JSON.parse(xhr.responseText));}
        };
        xhr.send();
    }""", "utf-8"))
        self.wfile.write(bytes("""
    function reloadLine(content) {{
        console.log(content);
        id= content["submodel"];
        console.log(id);
        dom = document.getElementById(id);
        console.log(dom);
        dom.innerHTML = ""
        dom.innerHTML += "<td>" + id + "</td>";
        results = content["properties"];
        console.log(results);
        for (prop of results) {{
            if (prop) {{
                dom.innerHTML += "<td><span class=\\"proved\\"></span></td>\\n";
            }} else {{
                dom.innerHTML += "<td><span class=\\"attack\\">X</span></td>\\n";
            }}
        }}
        dom.innerHTML += "<td><a href=javascript:proveSubmodel(\\"{}\\",\\""+id+"\\")>prove</a></td>";
        dom.innerHTML += "<td>"+ content["elapsed"] + "s</td>";
    }}""".format(model_name), "utf-8"))
        self.wfile.write(bytes("</script></body></html>", "utf-8"))

    def request_submodel(self, model, submodel):
        pass

    def get_submodel_latex(self, model, submodel):
        results_file = model.get_results_file(submodel)
        with open(results_file, "r") as fp:
            content = fp.read()
            parsed = json.loads(content)
            res = "{}".format(submodel)
            for i in range(len(parsed["properties"])):
                res += " & " + model.get_repr_attack(submodel, i)
            res += " & {:.2f}s".format(parsed["elapsed"])
            res += "\\\\ \\hline\n"
            return bytes(res, "utf-8")

    def get_submodel_html(self, model, submodel):
        results_file = model.get_results_file(submodel)
        if results_file is None:
            res = "<tr id=\"{}\">".format(submodel)
            res += "<td>{}</td>".format(submodel)
            res += "<td>N/A</td>"*len(model.get_properties())
            res += "<td><a href=javascript:proveSubmodel(\"{}\",\"{}\")>prove</a></td>".format(model.get_basename(), submodel)
            res += "<td>??? s</td>"
            res += "</tr>"
            return bytes(res, "utf-8")
        
        else:
            with open(results_file, "r") as fp:
                content = fp.read()
                parsed = json.loads(content)
                res = "<tr id=\"{}\">".format(submodel)
                res += "<td>{}</td>".format(submodel)
                for i in range(len(parsed["properties"])):
                    res += "<td>"+ model.get_repr_attack(submodel, i) + "</td>\n"
                res += "<td><a href=javascript:proveSubmodel(\"{}\",\"{}\")>prove</a></td>".format(model.get_basename(), submodel)
                res += "<td>{}s</td>".format(parsed["elapsed"])
                res += "</tr>"
                return bytes(res, "utf-8")


global webServer

if __name__ == "__main__":
    webServer = MyServer((hostName, serverPort), MyHandler)
    print("Server started http://{}:{}".format(hostName, serverPort))

    try:
        webServer.serve_forever()
    except KeyboardInterrupt:
        pass

    webServer.server_close()
    print("Server stopped.")
